from .shared import OPAYGOShared
from .shared_extended import OPAYGOSharedExtended


class OPAYGODecoder(object):
    MAX_TOKEN_JUMP = 64
    MAX_TOKEN_JUMP_COUNTER_SYNC = 100
    MAX_UNUSED_OLDER_TOKENS = 8*2

    @classmethod
    def get_activation_value_count_and_type_from_token(cls, token, starting_code, key, last_count,
                                                       restricted_digit_set=False, used_counts=None):
        if restricted_digit_set:
            token = OPAYGOShared.convert_from_4_digit_token(token)
        valid_older_token = False
        token_base = OPAYGOShared.get_token_base(token)  # We get the base of the token
        current_code = OPAYGOShared.put_base_in_token(starting_code, token_base)  # We put it into the starting code
        starting_code_base = OPAYGOShared.get_token_base(starting_code)  # We get the base of the starting code
        value = cls._decode_base(starting_code_base, token_base)  # If there is a match we get the value from the token
        # We try all combination up until last_count + TOKEN_JUMP, or to the larger jump if syncing counter
        # We could start directly the loop at the last count if we kept the token value for the last count
        if value == OPAYGOShared.COUNTER_SYNC_VALUE:
            max_count_try = last_count + cls.MAX_TOKEN_JUMP_COUNTER_SYNC + 1
        else:
            max_count_try = last_count + cls.MAX_TOKEN_JUMP + 1
        for count in range(0, max_count_try):
            masked_token = OPAYGOShared.put_base_in_token(current_code, token_base)
            if count % 2:
                this_type = OPAYGOShared.TOKEN_TYPE_SET_TIME
            else:
                this_type = OPAYGOShared.TOKEN_TYPE_ADD_TIME
            if masked_token == token:
                if cls._count_is_valid(count, last_count, value, this_type, used_counts):
                    return value, count, this_type
                else:
                    valid_older_token = True
            current_code = OPAYGOShared.generate_next_token(current_code, key)  # If not we go to the next token
        if valid_older_token:
            return -2, None, None
        return None, None, None

    @classmethod
    def _count_is_valid(cls, count, last_count, value, type, used_counts):
        if value == OPAYGOShared.COUNTER_SYNC_VALUE:
            if count > last_count - 30:
                return True
        elif count > last_count:
            return True
        elif cls.MAX_UNUSED_OLDER_TOKENS > 0:
            if count > last_count - cls.MAX_UNUSED_OLDER_TOKENS:
                if count not in used_counts and type == OPAYGOShared.TOKEN_TYPE_ADD_TIME:
                    return True
        return False

    @classmethod
    def update_used_counts(cls, past_used_counts, value, new_count, type):
        highest_count = max(past_used_counts) if past_used_counts else 0
        if new_count > highest_count:
            highest_count = new_count
        bottom_range = highest_count-cls.MAX_UNUSED_OLDER_TOKENS
        used_counts = []
        if type != OPAYGOShared.TOKEN_TYPE_ADD_TIME or value == OPAYGOShared.COUNTER_SYNC_VALUE or value == OPAYGOShared.PAYG_DISABLE_VALUE:
            # If it is not an Add-Time token, we mark all the past tokens as used in the range
            for count in range(bottom_range, highest_count+1):
                used_counts.append(count)
        else:
            # If it is an Add-Time token, we just mark the tokens actually used in the range
            for count in range(bottom_range, highest_count+1):
                if count == new_count or count in past_used_counts:
                    used_counts.append(count)
        return used_counts

    @classmethod
    def _decode_base(cls, starting_code_base, token_base):
        decoded_value = token_base - starting_code_base
        if decoded_value < 0:
            return decoded_value + 1000
        else:
            return decoded_value

    @classmethod
    def get_activation_value_count_from_extended_token(cls, token, starting_code, key, last_count,
                                                       restricted_digit_set=False):
        if restricted_digit_set:
            token = OPAYGOSharedExtended.convert_from_4_digit_token(token)
        token_base = OPAYGOSharedExtended.get_token_base(token) # We get the base of the token
        current_code = OPAYGOSharedExtended.put_base_in_token(starting_code, token_base) # We put it into the starting code
        starting_code_base = OPAYGOSharedExtended.get_token_base(starting_code) # We get the base of the starting code
        value = cls._decode_base_extended(starting_code_base, token_base)  # If there is a match we get the value from the token
        for count in range(0, 30):
            masked_token = OPAYGOSharedExtended.put_base_in_token(current_code, token_base)
            if masked_token == token and count > last_count:
                clean_count = count-1
                return value, clean_count
            current_code = OPAYGOSharedExtended.generate_next_token(current_code, key) # If not we go to the next token
        return None, None, None

    @classmethod
    def _decode_base_extended(cls, starting_code_base, token_base):
        decoded_value = token_base - starting_code_base
        if decoded_value < 0:
            return decoded_value + 1000000
        else:
            return decoded_value



from datetime import datetime
from ..encode_token import OPAYGOEncoder
from ..shared import OPAYGOShared


class SingleDeviceServerSimulator(object):

    def __init__(self, starting_code, key, starting_count=1, restricted_digit_set=False, time_divider=1):
        self.starting_code = starting_code
        self.key = key
        self.count = starting_count
        self.expiration_date = datetime.now()
        self.furthest_expiration_date = datetime.now()
        self.payg_enabled = True
        self.time_divider = time_divider
        self.restricted_digit_set = restricted_digit_set

    def print_status(self):
        print('Expiration Date: '+ str(self.expiration_date))
        print('Current count: '+str(self.count))
        print('PAYG Enabled: '+str(self.payg_enabled))

    def generate_payg_disable_token(self):
        self.count, token = OPAYGOEncoder.generate_standard_token(
            starting_code=self.starting_code,
            key=self.key,
            value=OPAYGOShared.PAYG_DISABLE_VALUE,
            count=self.count,
            restricted_digit_set=self.restricted_digit_set
        )
        return SingleDeviceServerSimulator._format_token(token)

    def generate_counter_sync_token(self):
        self.count, token = OPAYGOEncoder.generate_standard_token(
            starting_code=self.starting_code,
            key=self.key,
            value=OPAYGOShared.COUNTER_SYNC_VALUE,
            count=self.count,
            restricted_digit_set=self.restricted_digit_set
        )
        return SingleDeviceServerSimulator._format_token(token)

    def generate_token_from_date(self, new_expiration_date, force=False):
        furthest_expiration_date = self.furthest_expiration_date
        if new_expiration_date > self.furthest_expiration_date:
            self.furthest_expiration_date = new_expiration_date

        if new_expiration_date > furthest_expiration_date:
            # If the date is strictly above the furthest date activated, use ADD
            value = self._get_value_to_activate(new_expiration_date, self.expiration_date, force)
            self.expiration_date = new_expiration_date
            return self._generate_token_from_value(value, mode=OPAYGOShared.TOKEN_TYPE_ADD_TIME)
        else:
            # If the date is below or equal to the furthest date activated, use SET
            value = self._get_value_to_activate(new_expiration_date, datetime.now(), force)
            self.expiration_date = new_expiration_date
            return self._generate_token_from_value(value, mode=OPAYGOShared.TOKEN_TYPE_SET_TIME)

    def _generate_token_from_value(self, value, mode):
        self.count, token = OPAYGOEncoder.generate_standard_token(
            starting_code=self.starting_code,
            key=self.key,
            value=value,
            count=self.count,
            mode=mode,
            restricted_digit_set=self.restricted_digit_set
        )
        return SingleDeviceServerSimulator._format_token(token)

    def _generate_extended_value_token(self, value):
        pass

    def _get_value_to_activate(self, new_time, reference_time, force_maximum=False):
        if new_time <= reference_time:
            return 0
        else:
            days = self._timedelta_to_days(new_time - reference_time)
            value = int(round(days*self.time_divider, 0))
            if value > OPAYGOShared.MAX_ACTIVATION_VALUE:
                if not force_maximum:
                    raise Exception('TOO_MANY_DAYS_TO_ACTIVATE')
                else:
                    return OPAYGOShared.MAX_ACTIVATION_VALUE  # Will need to be activated again after those days
            return value

    @staticmethod
    def _timedelta_to_days(this_timedelta):
        return this_timedelta.days + (this_timedelta.seconds / 3600 / 24)

    @staticmethod
    def _format_token(token):
        token = str(token)
        if len(token) < 9:
            token = '0' * (9 - len(token)) + token
        return token



import siphash
import struct


class OPAYGOSharedExtended(object):
    MAX_BASE = 999999
    MAX_ACTIVATION_VALUE = 999999
    TOKEN_VALUE_OFFSET_EXTENDED = 1000000

    @classmethod
    def get_token_base(cls, code):
        return int(code % cls.TOKEN_VALUE_OFFSET_EXTENDED)

    @classmethod
    def put_base_in_token(cls, token, token_base):
        if token_base > cls.MAX_BASE:
            Exception('INVALID_VALUE')
        return token - cls.get_token_base(token) + token_base

    @classmethod
    def generate_next_token(cls, last_code, key):
        conformed_token = struct.pack('>Q', last_code) # We convert the token to bytes
        token_hash = siphash.SipHash_2_4(key, conformed_token).hash() # We hash it
        new_token = cls._convert_hash_to_token(token_hash) # We convert to token and return
        return new_token

    @classmethod
    def _convert_hash_to_token(cls, this_hash):
        token = cls._convert_to_40_bits(this_hash) # We convert the 64bits value to an INT no greater than 12 digits
        return token

    @classmethod
    def _convert_to_40_bits(cls, source):
        mask = ((1 << (64 - 24 + 1)) - 1) << 24
        temp = (source & mask) >> 24
        if temp > 999999999999:
            temp = temp - 99511627777
        return temp

    @classmethod
    def convert_to_4_digit_token(cls, source):
        restricted_digit_token = ''
        bit_array = cls._bit_array_from_int(source, 40)
        for i in range(20):
            this_array = bit_array[i*2:(i*2)+2]
            restricted_digit_token += str(cls._bit_array_to_int(this_array)+1)
        return int(restricted_digit_token)

    @classmethod
    def convert_from_4_digit_token(cls, source):
        bit_array = []
        for digit in str(source):
            digit = int(digit) - 1
            this_array = cls._bit_array_from_int(digit, 2)
            bit_array += this_array
        return cls._bit_array_to_int(bit_array)

    @classmethod
    def _bit_array_to_int(cls, bit_array):
        integer = 0
        for bit in bit_array:
            integer = (integer << 1) | bit
        return integer

    @classmethod
    def _bit_array_from_int(cls, source, bits):
        bit_array = []
        for i in range(bits):
            bit_array += [bool(source & (1 << (bits - 1 - i)))]
        return bit_array



from helpers import ADD_TIME, SET_TIME, DISABLE_VALUE, generate_from_device_data, test_accepted, test_how_many_days, test_name


def run_core_token_tests(device_data):

    test = 'We enter an invalid token'
    token_g1 = '123 456 789'
    test_accepted('G1', token_g1, False, description=test)

    test = 'We enter a valid token for setting one day'
    device_data, token_g2 = generate_from_device_data(device_data, token_type=SET_TIME, value_days=1)
    test_how_many_days('G2', token_g2, 1, description=test)

    test = 'We enter a valid token for adding one day'
    device_data, token_g3 = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=1)
    test_how_many_days('G3', token_g3, 2, description=test)

    test = 'We enter the same Add Time token for 1 day, the days should not be added and the device should signal that the token was already used'
    test_how_many_days('G4A', token_g3, 2, description=test)
    test = 'We enter the older Set Time token for 1 day, the days should not change and the device should signal that the token was already used'
    test_how_many_days('G4B', token_g2, 2, description=test)

    test = 'We enter a valid token for setting 30 days and ensures it sets and does not add to the existing'
    device_data, token_g5 = generate_from_device_data(device_data, token_type=SET_TIME, value_days=30)
    test_how_many_days('G5', token_g5, 30, description=test)

    test = 'We enter a valid token for setting 0 days and ensures the device is inactive with the outputs disabled immediately'
    device_data, token_g6 = generate_from_device_data(device_data, token_type=SET_TIME, value_days=0)
    test_how_many_days('G6', token_g6, 0, description=test)

    test = 'We enter 3 consecutive Add Time tokens with the maximum amount of days and ensure that they cumulate properly'
    for i in range(1,3+1):
        device_data, token_g7 = generate_from_device_data(device_data, token_type=ADD_TIME, value_raw=995)
        test_how_many_days(test_name('G7', i), token_g7, value_raw=995*i, device_data=device_data, description=test)
        test = ''

    test = 'We enter 21 consecutive Set Time tokens for 1, 2, 3, … 21 days each with a count 30 higher than the other. The validation of the token should not take more than 5 seconds'
    for i in range(1,21+1):
        device_data, token_g8 = generate_from_device_data(device_data, token_type=SET_TIME, value_days=i, token_count=device_data['token_count']+29)
        test_how_many_days(test_name('G8', i), token_g8, value_days=i, device_data=device_data, description=test)
        test = ''

    test = 'We enter a PAYG Disable token into the device'
    device_data, token_g9 = generate_from_device_data(device_data, token_type=SET_TIME, value_raw=DISABLE_VALUE)
    test_how_many_days('G9', token_g9, None, description=test)

    test = 'We enter a Set Time token for 0 day, it should relock the device'
    device_data, token_g10 = generate_from_device_data(device_data, token_type=SET_TIME, value_days=0)
    test_how_many_days('G10', token_g10, 0, description=test)

    test = 'We enter a PAYG Disable token to relock the device, then enter a Add Time token with 0 day, it should NOT relock the device (Optional)'
    device_data, token_g11a = generate_from_device_data(device_data, token_type=SET_TIME, value_raw=DISABLE_VALUE)
    test_how_many_days('G11A', token_g11a, None, description=test)
    device_data, token_g11b = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=0)
    test_how_many_days('G11B', token_g11b, None)

    test = 'We deactivate the device with a Set Time of 0 days. We then wait 48 hours before entering a Add Time of 1 day and ensuring that the days late are not considered in the activation time'
    device_data, token_g12a = generate_from_device_data(device_data, token_type=SET_TIME, value_days=0)
    test_how_many_days('G12A', token_g12a, 0, description=test)
    device_data, token_g12b = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=1)
    test_how_many_days('G12B', token_g12b, 1)

    return device_data


def run_unordered_entry_tests(device_data):

    test = 'We generate 3 Add Time tokens, then enter the 3rd, then first, then second and ensure the days are added properly'
    device_data, token_u1a = generate_from_device_data(device_data, token_type=SET_TIME, value_days=60)
    test_how_many_days('U1A', token_u1a, 60, description=test)
    device_data, token_u1b = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=10)
    device_data, token_u1c = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=5)
    device_data, token_u1d = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=1)
    test_how_many_days('U1B', token_u1d, 61)
    test_how_many_days('U1C', token_u1b, 71)
    test_how_many_days('U1D', token_u1c, 76)

    test = 'We generate an Add Time, a Set Time and another Add Time token. We enter the set time and ensure that the older Add Time does not work but the newer does'
    device_data, token_u2a = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=5)
    device_data, token_u2b = generate_from_device_data(device_data, token_type=SET_TIME, value_days=10)
    device_data, token_u2c = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=3)
    test_how_many_days('U2A', token_u2b, 10, description=test)
    test_accepted('U2B', token_u2a, False)
    test_how_many_days('U2C', token_u2c, 13)

    test = 'We generate an Add Time token and a Disable PAYG token, we enter the Disable PAYG token and then the Add Time token should be refused'
    device_data, token_u3a = generate_from_device_data(device_data, token_type=ADD_TIME, value_days=1)
    device_data, token_u3b = generate_from_device_data(device_data, token_type=SET_TIME, value_raw=DISABLE_VALUE)
    test_how_many_days('U3A', token_u3b, None, description=test)
    test_accepted('U3B', token_u3a, False)

    return device_data




if __name__ == '__main__':
    device_data = {
        'serial_number': 'changeme',
        'starting_code': 123456789,
        'key': 'a29ab82edc5fbbc41ec9530f6dac86b1',
        'restricted_digit_set': False,
        'time_divider': 1,
        'token_count': 1
    }
    print('Test Name,Token Used,Expected Result,Test Details,Result')
    device_data = run_core_token_tests(device_data)
    device_data = run_unordered_entry_tests(device_data)



from datetime import datetime, timedelta

from importlib import import_module
openpaygo_token = import_module("openpaygo-token")
simulators = import_module("openpaygo-token.simulators")


def assert_time_equals(time1, time2):
    assert time1.replace(second=0, microsecond=0) == time2.replace(second=0, microsecond=0)


if __name__ == '__main__':
    # ------ IMPORTANT WARNING --------
    # DO NOT USE THIS KEY IN PRODUCTION
    # IT IS JUST AN EXAMPLE
    # ---------------------------------
    device_key = b'\xa2\x9a\xb8.\xdc_\xbb\xc4\x1e\xc9S\x0fm\xac\x86\xb1'
    device_starting_code = 123456789 # Generated by fair dice roll
    restricted_digit_set = False

    print('Device: We initiate the device simulator with our device')
    device_simulator = simulators.DeviceSimulator(device_starting_code, device_key,
                                       restricted_digit_set=restricted_digit_set,
                                       waiting_period_enabled=False)
    print('Server: We initiate the server simulator with our device')
    server_simulator = simulators.SingleDeviceServerSimulator(device_starting_code, device_key,
                                                   restricted_digit_set=restricted_digit_set)

    print('\n')
    print('Device: We try entering an invalid token into the device: 123456789')
    device_simulator.enter_token('123456789')
    print('Device: We check the device status (should be still inactive)')
    device_simulator.print_status()

    print('\n')
    print('Server: We add 1 days of activation for the device')
    this_token = server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    print('Token: '+this_token)
    print('Device: We enter the generated token into the device')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should be active with 1 day)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=1))

    print('\n')
    print('Device: We enter the token a second time to make sure it doesnt add the days again')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should be active with 1 day)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=1))

    print('\n')
    print('Server: We set it to expire in 30 days')
    this_token = server_simulator.generate_token_from_date(datetime.now() + timedelta(days=30))
    print('Token: ' + this_token)
    print('Device: We enter the generated token into the device')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should be active with 30 days)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=30))

    print('\n')
    print('Server: We set it to expire in 7 days (removing 23 days)')
    this_token = server_simulator.generate_token_from_date(datetime.now() + timedelta(days=7))
    print('Token: ' + this_token)
    print('Device: We enter the generated token into the device')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should be active with 7 days)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=7))

    print('\n')
    print('Server: We generate a token for putting the device in PAYG-OFF mode')
    this_PAYG_OFF_code = server_simulator.generate_payg_disable_token()
    this_PAYG_OFF_code = str(this_PAYG_OFF_code)
    print('Token: ' + this_PAYG_OFF_code)
    print('Device: We enter the generated token into the device')
    device_simulator.enter_token(this_PAYG_OFF_code)
    print('Device: We check the device status (should be active forver)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert device_simulator.payg_enabled is False

    print('\n')
    print('Server: We generate a token for putting the device back PAYG-ON mode with 0 days')
    this_token = server_simulator.generate_token_from_date(datetime.now() + timedelta(days=0))
    this_token = str(this_token)
    print('Token: ' + this_token)
    print('Device: We enter the generated token into the device')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should not be active)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert device_simulator.payg_enabled is True
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=0))

    print('\n')
    print('Server: We generate a bunch of 1 day tokens, but only enter the latest one')
    server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    this_token = server_simulator.generate_token_from_date(datetime.now() + timedelta(days=1))
    this_token = str(this_token)
    print('Token: ' + this_token)
    print('Device: We enter the latest generated token into the device')
    device_simulator.enter_token(this_token)
    print('Device: We check the device status (should be active with 1 day and the count synchronised with the server)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=1))

    print('\n')
    print('Server: We add generate 9 tokens each add-time of 1 day')
    token_1 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_2 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_3 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_4 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_5 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_6 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_7 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_8 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_9 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    print('Tokens: ', token_1, token_2, token_3, token_4, token_5, token_6, token_7, token_8, token_9)
    print('Device: We enter the 9th token into the device')
    device_simulator.enter_token(token_9)
    print('Device: We check the device status (should be active with +1 day (2 days total)')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=2))
    print('Device: We enter the 1st token into the device')
    device_simulator.enter_token(token_1)
    print('Device: We check the device status , it should not have changed, because its more than 5 add times before')
    device_simulator.print_status()
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=2))
    print('Device: We enter the tokens 5, 4, 3 and 2 into the device')
    device_simulator.enter_token(token_5)
    device_simulator.enter_token(token_4)
    device_simulator.enter_token(token_3)
    device_simulator.enter_token(token_2)
    print('Device: We check the device status , it should have +4 days (6 days total)')
    device_simulator.print_status()
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=6))

    print('\n')
    print('Server: We add generate 2 tokens, first add-time and then set-time')
    token_1 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    token_2 = server_simulator._generate_token_from_value(0, openpaygo_token.OPAYGOShared.TOKEN_TYPE_SET_TIME)
    print('Tokens: ', token_1, token_2)
    print('Device: We enter the 2nd token')
    device_simulator.enter_token(token_2)
    print('Device: We check the device status should be active in 0 days.')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=0))
    print('Device: We enter the 1st token into the device')
    device_simulator.enter_token(token_1)
    print('Device: We check the device status , it should not have changed, '
          'because you cannot use an add-time token older than a set-time')
    device_simulator.print_status()
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=0))

    print('\n')
    print('Server: We add generate 2 tokens, first add-time and then set-time')
    token_1 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_SET_TIME)
    token_2 = server_simulator._generate_token_from_value(1, openpaygo_token.OPAYGOShared.TOKEN_TYPE_ADD_TIME)
    print('Tokens: ', token_1, token_2)
    print('Device: We enter the 2nd token')
    device_simulator.enter_token(token_2)
    print('Device: We check the device status should be active in 0 days.')
    device_simulator.print_status()
    assert device_simulator.count == server_simulator.count
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=1))
    print('Device: We enter the 1st token into the device')
    device_simulator.enter_token(token_1)
    print('Device: We check the device status , it should not have changed, '
          'because you cannot use an older set-time token')
    device_simulator.print_status()
    assert_time_equals(device_simulator.expiration_date, datetime.now() + timedelta(days=1))



import codecs

from importlib import import_module
openpaygo_token = import_module("openpaygo-token")


SET_TIME = 1
ADD_TIME = 2
DISABLE_VALUE = 998

START_RED = '\033[91m'
END_RED = '\033[0m'


def generate_from_device_data(device_data, token_type, value_raw=None, value_days=None, token_count=None):
    assert (value_days is not None) or (value_raw is not None)
    if value_raw is None:
        value_raw = value_days*device_data['time_divider']
    device_data['token_count'], token = openpaygo_token.OPAYGOEncoder.generate_standard_token(
        starting_code=device_data['starting_code'],
        key=codecs.decode(device_data['key'], 'hex'),
        value=value_raw,
        count=token_count or device_data['token_count'],
        restricted_digit_set=device_data['restricted_digit_set'],
        mode=token_type
    )
    token = str(token).rjust(9, '0')
    token = ' '.join([token[i:i + 3] for i in range(0, len(token), 3)])
    return device_data, token


def test_how_many_days(test_name, token, value_days=None, value_raw=None, device_data=None, description=''):
    if value_days is None:
        if value_raw is not None:
            value_days = value_raw/device_data['time_divider']
        else:
            value_days = 'infinite'
    print(test_name+','+token+','+str(value_days)+' days active'+',"'+description+'"')


def test_accepted(test_name, token, expected, description=''):
    expected_string = 'Token Accepted' if expected else 'Token Refused'
    print(test_name+','+token+','+expected_string+',"'+description+'"')


def test_how_many_days_validator(device_simulator, test_name, token, value_days=None, value_raw=None, device_data=None, description='', accepted_but_used=False):
    if value_days is None:
        if value_raw is not None:
            value_days = value_raw/device_data['time_divider']
        else:
            value_days = 'infinite'
    result = device_simulator.enter_token(token.replace(' ', ''), show_result=False)
    if result == 1 or (accepted_but_used and result == -2):
        if device_simulator.get_days_remaining() == value_days:
            print(test_name+': Passed')
            return
    print(START_RED+test_name+': Failed'+END_RED)


def test_accepted_validator(device_simulator, test_name, token, expected, description=''):
    if device_simulator.enter_token(token.replace(' ', ''), show_result=False) == expected:
        print(test_name+': Passed')
        return
    else:
        print(START_RED+test_name+': Failed'+END_RED)


def test_name(test_base_name, test_number):
    return test_base_name+chr(test_number+64)



from datetime import datetime, timedelta
from ..shared import OPAYGOShared
from ..decode_token import OPAYGODecoder


class DeviceSimulator(object):

    def __init__(self, starting_code, key, starting_count=1, restricted_digit_set=False, waiting_period_enabled=True, time_divider=1):
        self.starting_code = starting_code
        self.key = key
        self.time_divider = time_divider
        self.restricted_digit_set = restricted_digit_set
        self.waiting_period_enabled = waiting_period_enabled  # Should always be true except for testing

        self.payg_enabled = True
        self.count = starting_count
        self.expiration_date = datetime.now()
        self.invalid_token_count = 0
        self.token_entry_blocked_until = datetime.now()
        self.used_counts = []

    def print_status(self):
        print('-------------------------')
        print('Expiration Date: '+ str(self.expiration_date))
        print('Current count: '+str(self.count))
        print('PAYG Enabled: '+str(self.payg_enabled))
        print('Active: '+str(self.is_active()))
        print('-------------------------')

    def is_active(self):
        return self.expiration_date > datetime.now()

    def enter_token(self, token, show_result=True):
        if len(token) == 9:
            token_int = int(token)
            return self._update_device_status_from_token(token_int, show_result)
        else:
            token_int = int(token)
            return self._update_device_status_from_extended_token(token_int, show_result)

    def get_days_remaining(self):
        if self.payg_enabled:
            td = self.expiration_date - datetime.now()
            days, hours, minutes = td.days, td.seconds//3600, (td.seconds//60)%60
            days = days + (hours + minutes/60)/24
            return round(days)
        else:
            return 'infinite'

    def _update_device_status_from_token(self, token, show_result=True):
        if self.token_entry_blocked_until > datetime.now() and self.waiting_period_enabled:
            if show_result:
                print('TOKEN_ENTRY_BLOCKED')
            return False
        token_value, token_count, token_type = OPAYGODecoder.get_activation_value_count_and_type_from_token(
            token=token,
            starting_code=self.starting_code,
            key=self.key,
            last_count=self.count,
            restricted_digit_set=self.restricted_digit_set,
            used_counts=self.used_counts
        )
        if token_value is None:
            if show_result:
                print('TOKEN_INVALID')
            self.invalid_token_count += 1
            self.token_entry_blocked_until = datetime.now() + timedelta(minutes=2**self.invalid_token_count)
            return -1
        elif token_value == -2:
            if show_result:
                print('OLD_TOKEN')
            return -2
        else:
            if show_result:
                print('TOKEN_VALID', ' | Value:', token_value)
            if token_count > self.count or token_value == OPAYGOShared.COUNTER_SYNC_VALUE:
                self.count = token_count
            self.used_counts = OPAYGODecoder.update_used_counts(self.used_counts, token_value, token_count, token_type)
            self.invalid_token_count = 0
            self._update_device_status_from_token_value(token_value, token_type)
            return 1

    def _update_device_status_from_extended_token(self, token):
        if self.token_entry_blocked_until > datetime.now() and self.waiting_period_enabled:
            print('TOKEN_ENTRY_BLOCKED')
        token_value, token_count = OPAYGODecoder.get_activation_value_count_from_extended_token(
            token=token,
            starting_code=self.starting_code,
            key=self.key,
            last_count=self.count,
            restricted_digit_set=self.restricted_digit_set
        )
        if token_value is None:
            print('TOKEN_INVALID')
            self.invalid_token_count += 1
            self.token_entry_blocked_until = datetime.now() + timedelta(minutes=2**self.invalid_token_count)
        else:
            print('Special token entered, value: '+str(token_value))

    def _update_device_status_from_token_value(self, token_value, token_type):
        if token_value <= OPAYGOShared.MAX_ACTIVATION_VALUE:
            if not self.payg_enabled and token_type == OPAYGOShared.TOKEN_TYPE_SET_TIME:
                self.payg_enabled = True
            if self.payg_enabled:
                self._update_expiration_date_from_value(token_value, token_type)
        elif token_value == OPAYGOShared.PAYG_DISABLE_VALUE:
            self.payg_enabled = False
        elif token_value != OPAYGOShared.COUNTER_SYNC_VALUE:
            # We do nothing if its the sync counter value, the counter has been synced already
            print('COUNTER_SYNCED')
        else:
            # If it's another value we also do nothing, as they are not defined
            print('UNKNOWN_COMMAND')

    def _update_expiration_date_from_value(self, toke_value, token_type):
        number_of_days = toke_value/self.time_divider
        if token_type == OPAYGOShared.TOKEN_TYPE_SET_TIME:
            self.expiration_date = datetime.now() + timedelta(days=number_of_days)
        else:
            if self.expiration_date < datetime.now():
                self.expiration_date = datetime.now()
            self.expiration_date = self.expiration_date + timedelta(days=number_of_days)



from .shared import OPAYGOShared
from .shared_extended import OPAYGOSharedExtended


class OPAYGOEncoder(object):

    @classmethod
    def generate_standard_token(cls, starting_code, key, value, count,
                                mode=OPAYGOShared.TOKEN_TYPE_SET_TIME, restricted_digit_set=False):
        # We get the first 3 digits with encoded value
        starting_code_base = OPAYGOShared.get_token_base(starting_code)
        token_base = cls._encode_base(starting_code_base, value)
        current_token = OPAYGOShared.put_base_in_token(starting_code, token_base)
        current_count_odd = count % 2
        if mode == OPAYGOShared.TOKEN_TYPE_SET_TIME:
            if current_count_odd: # Odd numbers are for Set Time
                new_count = count+2
            else:
                new_count = count+1
        else:
            if current_count_odd: # Even numbers are for Add Time
                new_count = count+1
            else:
                new_count = count+2
        for xn in range(0, new_count):
            current_token = OPAYGOShared.generate_next_token(current_token, key)
        final_token = OPAYGOShared.put_base_in_token(current_token, token_base)
        if restricted_digit_set:
            final_token = OPAYGOShared.convert_to_4_digit_token(final_token)
            final_token = '{:015d}'.format(final_token)
        else:
            final_token = '{:09d}'.format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base(cls, base, number):
        if number + base > 999:
            return number + base - 1000
        else:
            return number + base

    @classmethod
    def generate_extended_token(cls, starting_code, key, value, count, restricted_digit_set=False):
        starting_code_base = OPAYGOSharedExtended.get_token_base(starting_code)
        token_base = cls._encode_base_extended(starting_code_base, value)
        current_token = OPAYGOSharedExtended.put_base_in_token(starting_code, token_base)
        new_count = count + 1
        for xn in range(0, new_count):
            current_token = OPAYGOSharedExtended.generate_next_token(current_token, key)
        final_token = OPAYGOSharedExtended.put_base_in_token(current_token, token_base)
        if restricted_digit_set:
            final_token = OPAYGOSharedExtended.convert_to_4_digit_token(final_token)
            final_token = '{:020d}'.format(final_token)
        else:
            final_token = '{:012d}'.format(final_token)
        return new_count, final_token

    @classmethod
    def _encode_base_extended(cls, base, number):
        if number + base > 999999:
            return number + base - 1000000
        else:
            return number + base
